# Copyright (C) 2023 Wibo Kuipers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Literal, List, Tuple

DatabaseColumnTypeIntVariations = Literal["INT", "TINYINT", "BIGINT"]
DatabaseColumnTypeTextVariations = Literal["TEXT", "LONGTEXT"]
DatabaseColumnType = Literal[
    DatabaseColumnTypeIntVariations,
    DatabaseColumnTypeTextVariations,
    "FLOAT",
    "DECIMAL",
    "VARCHAR",
    "DATETIME",
    "JSON",
]

Method = Literal["createtable"] | Literal["altertable"]
Method.__doc__ = "Available version source methods"

ValidationIssueLevel = Literal["LOW"] | Literal["HIGH"] | Literal["CRITICAL"]
ValidationIssueLevel.__doc__ = """
LOW: Will work, but will not have any effect on the database.
HIGH: Will still work, but might produce a different result than desired.
CRITICAL: Will not execute.
"""
ValidationIssuesList = List[Tuple[ValidationIssueLevel, str]]
ValidationIssuesList.__doc__ = "An issue is structured like Tuple(level, issue)"



import json

import pytest

from alphadb import AlphaDB
from alphadb.utils.exceptions import NeedsConfirmation

db = AlphaDB()
db.connect(host="localhost", user="root", password="test", database="test")


#### Init method
def test_mysql_init():
    init = db.init()
    assert init == True
    init = db.init()
    assert init == "already-initialized"


#### Status method
def test_mysql_status():
    status = db.status()
    assert status == {"init": True, "version": "0.0.0", "name": "test", "template": None}


#### Update method
def test_mysql_update():
    with open("../tests/assets/test-db-structure.json") as f:
        structure = json.loads(f.read())

    assert db.update(version_source=structure) == True


#### Vacate method
def test_mysql_vacate():
    #### Confirm not specified
    with pytest.raises(NeedsConfirmation):
        db.vacate()

    assert db.vacate(confirm=True) == True



# Copyright (C) 2023 Wibo Kuipers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import json
import os

from cli.utils.config import config_get, config_write, config_get_items
from requests import get
from requests.exceptions import JSONDecodeError, ConnectionError
from inquirer import Text, prompt, List, Path
from inquirer.errors import ValidationError
from cli.utils.common import console, clear

def select_version_source() -> str | None:
    "Get version sources from config and promt user to select one"
    version_sources = config_get_items("VERSION_SOURCES")
    version_source_path = None

    #### If no versions sources exist, ask to create one
    if len(version_sources) == 0:
        console.print("[cyan]You have no saved version sources, so let's find one.[/cyan]\n")
        version_source_path = add_version_source()
        print("\n")
     
    #### Ask user to select version source
    if version_source_path == None:

        choices = [f"{i[0]} ({'web' if i[1][0:4] == 'http' else 'file'})" for i in version_sources]
        choices.append("+ New version source")
        questions = [List("version_source", message=f"Found {len(version_sources)} version sources, which one do you wish to use?", choices=choices)]

        answers = prompt(questions)

        #### If answers is None, user probably aborted
        if answers == None: return

        version_source = answers["version_source"]

        clear()

        if version_source == "+ New version source":
            version_source_path = add_version_source()
        else:
            version_source_path = config_get("VERSION_SOURCES")[version_source.split(" ")[0]]

        return version_source_path


def add_version_source() -> str | None:
    "Adding a new version source to config"


    print(
        "Version sources can either be local JSON files or URL's returning JSON data.\n"
    )

    try:  ## To catch ConnectionError


        def pathval(path: str):
            if path[0:4] == "http":
                r = get(path)
                if not r.status_code == 200:
                    raise ValidationError(
                        "", reason=f"URL responded with {r.status_code}"
                    )
                return True

            if not os.path.isfile(path):
                raise ValidationError(
                    "", reason="This path does not point towards a JSON file."
                )
            return True

        questions = [
            Path(
                "path",
                message="Please provide either an absolute local file path or a local/remote URL",
                validate=lambda _, x: pathval(x),
            )
        ]
        answers = prompt(questions)

        #### If answers is None, user probably aborted
        if answers == None:
            return

        path = answers["path"]

        if path[0:4] == "http":
            template_name = get(path).json()["name"]
        else:
            with open(path) as version_json:
                template_name = json.load(version_json)["name"]

    #### TODO This connection error catch does not seem to work
    except ConnectionError:
        console.print("\n[red]Unable to establish connection[/red]\n")
        return

    except JSONDecodeError:
        console.print(
            "\n[red]The supplied URL did not respond with compatible data[/red]\n"
        )
        return

    ##### TODO maybe add a check if the data has no errors

    questions = [Text("name", message="Name the template", default=template_name)]
    answers = prompt(questions)

    #### If answers is None, user probably aborted
    if answers == None:
        return

    name = answers["name"]

    vs = config_get("VERSION_SOURCES", fallback=True)
    if vs == None:
        vs = {}
    vs[name] = path
    config_write({"VERSION_SOURCES": vs})

    return path

def vs_path_to_json(path) -> dict | None:
    
    #### Get version invormation from path
    if path[0:4] == "http":
        try:
            r = get(path)
            if not r.status_code == 200:
                console.print(f"[red]No version information was returned from the server, it instead responded with status {r.status_code}[/red]\n")
                return


            return r.json()

        #### TODO This connection error catch does not seem to work  
        except ConnectionError:
            console.print("[red]Unable to establish connection[/red]\n")
            return

        except JSONDecodeError:
            console.print("[red]The supplied URL did not respond with compatible data[/red]\n")
            return
    else:
        with open(path) as version_json:
            return json.load(version_json)



# Copyright (C) 2023 Wibo Kuipers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from alphadb.utils.exceptions import IncompleteVersionObject
from alphadb.utils.query.column.definecolumn import definecolumn, prepare_definecolumn_data


def createtable(version_source: dict, table_name: str, version: str):
    #### Define query base
    query = f" CREATE TABLE {table_name} ("

    #### Get the data for the current table
    table_data = next(v["createtable"][table_name] for v in version_source["version"] if v["_id"] == version)

    #### Loop through table columns
    for column in table_data:
        #### Using method `addcolumn` as the createtable column definition is identical to the addcolumn submethod of altertable
        column_data = prepare_definecolumn_data(table_name=table_name, column=column, table_data=table_data, version=version)

        #### If column data is null, its some attribute that should be handled later (foreign_key, primary_key, etc...)
        if column_data == None:
            continue

        #### Create query chunk
        query += definecolumn(
            column_name=column,
            column_type=table_data[column]["type"],
            length=column_data["length"],
            null=column_data["null"],
            unique=column_data["unique"],
            default=column_data["default"],
            auto_increment=column_data["auto_increment"],
        )

        #### Add a comma (trailing comma will be removed after loop, less complex logic)
        query += ","

    #### Remove trailing comma
    query = query[:-1]

    #### Primary key
    if "primary_key" in table_data:
        query += f', PRIMARY KEY ({table_data["primary_key"]})'

    #### Foreign key
    if "foreign_key" in table_data:
        foreign_key = table_data["foreign_key"]  ## Just for readability
        if not "key" in foreign_key:
            raise IncompleteVersionObject(key="key", object="foreign_key")

        if not "references" in foreign_key:
            raise IncompleteVersionObject(key="references", object="foreign_key")

        if "on_delete" in foreign_key:
            query += f', FOREIGN KEY ({foreign_key["key"]}) REFERENCES {foreign_key["references"]} ({foreign_key["key"]}) ON DELETE CASCADE'

    #### End of query
    query += " ) ENGINE = InnoDB;"

    return query



class UnexpectedError(Exception):
    msg = "An unexpected error has occured."
    def __init__(self):
        super().__init__(self, self.msg)

class ConfigIncoplete(Exception):
    msg = "Config seems to be incomplete. The config file might be broken or edited."
    def __init__(self):
        super().__init__(self, self.msg)



# Copyright (C) 2023 Wibo Kuipers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from sys import platform
import os
from pathlib import Path
from cli.utils.types import AlphaDBMock

global CONFIG_PATH
global db
db = AlphaDBMock()

if platform == "linux" or platform == "linux2" or platform == "darwin":
    config_dir = os.path.join(str(Path.home()), ".config/alphadb/")
    if not os.path.exists(config_dir): os.mkdir(config_dir)
    CONFIG_PATH = os.path.join(config_dir, "cli-config.ini")
else:
    CONFIG_PATH = "config.ini"




# Copyright (C) 2023 Wibo Kuipers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.


from collections.abc import Callable

from .exceptions import DBNotInitialized, NoConnection


#### Checks for database connection
def conn_test(func: Callable) -> Callable:
    def _(inst, *args, **kwargs):
        if inst.connection == None:
            raise NoConnection()  ## Raise
        else:
            return func(inst, *args, **kwargs)

    return _


#### Wraps around check function, if db is not initialized some functions should not be called
def init_test(func: Callable) -> Callable:
    def _(self, *args, **kwargs):
        if self.check()["check"]:
            return func(self, *args, **kwargs)
        else:
            raise DBNotInitialized()

    return _
